# __init__.py dentro de algoritmos/
from .alg_bin import cl_alg_stn_bin
from .alg_quantum import cl_alg_quantum
from .alg_eda import cl_alg_eda
from .alg_gp import cl_alg_gp

__all__ = ["cl_alg_stn_bin", "cl_alg_quantum", "cl_alg_eda", "cl_alg_gp"]
